@@ -47,6 +47,10 @@
 #include <pthread.h>
 #include <errno.h>
 
+#if defined HAVE_XLOCALE_HEADER
+#include <xlocale.h>
+#endif
+
 #if defined HAVE_POSIX_SPAWN || defined HAVE_POSIX_SPAWNP
 #include <spawn.h>
 #endif
@@ -68,32 +72,40 @@ extern char **environ;
 # define ENV_SIZE 2
 #endif
 
-#define DLSYM(TYPE_, VAR_, SYMBOL_)                                            \
-    union {                                                                    \
-        void *from;                                                            \
-        TYPE_ to;                                                              \
-    } cast;                                                                    \
-    if (0 == (cast.from = dlsym(RTLD_NEXT, SYMBOL_))) {                        \
-        perror("bear: dlsym");                                                 \
-        exit(EXIT_FAILURE);                                                    \
-    }                                                                          \
+#define STRINGIFY(x) #x
+#define TOSTRING(x) STRINGIFY(x)
+#define AT "libear: (" __FILE__ ":" TOSTRING(__LINE__) ") "
+
+#define PERROR(msg) do { perror(AT msg); } while (0)
+
+#define ERROR_AND_EXIT(msg) do { PERROR(msg); exit(EXIT_FAILURE); } while (0)
+
+#define DLSYM(TYPE_, VAR_, SYMBOL_)                                 \
+    union {                                                         \
+        void *from;                                                 \
+        TYPE_ to;                                                   \
+    } cast;                                                         \
+    if (0 == (cast.from = dlsym(RTLD_NEXT, SYMBOL_))) {             \
+        PERROR("dlsym");                                            \
+        exit(EXIT_FAILURE);                                         \
+    }                                                               \
     TYPE_ const VAR_ = cast.to;
 
 
 typedef char const * bear_env_t[ENV_SIZE];
 
-static int bear_capture_env_t(bear_env_t *env);
-static void bear_release_env_t(bear_env_t *env);
-static char const **bear_update_environment(char *const envp[], bear_env_t *env);
-static char const **bear_update_environ(char const **in, char const *key, char const *value);
-static void bear_report_call(char const *const argv[]);
-static int bear_write_json_report(int fd, char const *const cmd[], char const *cwd, pid_t pid);
-static int bear_encode_json_string(char const *src, char *dst, size_t dst_size);
-static char const **bear_strings_build(char const *arg, va_list *ap);
-static char const **bear_strings_copy(char const **const in);
-static char const **bear_strings_append(char const **in, char const *e);
-static size_t bear_strings_length(char const *const *in);
-static void bear_strings_release(char const **);
+static int capture_env_t(bear_env_t *env);
+static void release_env_t(bear_env_t *env);
+static char const **string_array_partial_update(char *const envp[], bear_env_t *env);
+static char const **string_array_single_update(char const **in, char const *key, char const *value);
+static void report_call(char const *const argv[]);
+static void write_report(int fd, char const *const argv[]);
+static int write_json_report(int fd, char const *const cmd[], char const *cwd, pid_t pid);
+static int encode_json_string(char const *src, char *dst, size_t dst_size);
+static char const **string_array_from_varargs(char const *arg, va_list *ap);
+static char const **string_array_copy(char const **const in);
+static size_t string_array_length(char const *const *in);
+static void string_array_release(char const **);
 
 
 static bear_env_t env_names =
@@ -114,10 +126,14 @@ static bear_env_t initial_env =
 
 static int initialized = 0;
 static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
+static locale_t utf_locale;
 
 static void on_load(void) __attribute__((constructor));
 static void on_unload(void) __attribute__((destructor));
 
+static int mt_safe_on_load(void);
+static void mt_safe_on_unload(void);
+
 
 #ifdef HAVE_EXECVE
 static int call_execve(const char *path, char *const argv[],
@@ -159,28 +175,50 @@ static int call_posix_spawnp(pid_t *restrict pid, const char *restrict file,
 
 static void on_load(void) {
     pthread_mutex_lock(&mutex);
-#ifdef HAVE_NSGETENVIRON
-    environ = *_NSGetEnviron();
-#endif
-    if (!initialized)
-        initialized = bear_capture_env_t(&initial_env);
+    if ((!initialized) && (mt_safe_on_load()))
+        initialized = 1;
     pthread_mutex_unlock(&mutex);
 }
 
 static void on_unload(void) {
     pthread_mutex_lock(&mutex);
-    bear_release_env_t(&initial_env);
+    if (initialized)
+        mt_safe_on_unload();
     initialized = 0;
     pthread_mutex_unlock(&mutex);
 }
 
+static int mt_safe_on_load(void) {
+#ifdef HAVE_NSGETENVIRON
+    environ = *_NSGetEnviron();
+    if (0 == environ)
+        return 0;
+#endif
+    // Create locale to encode UTF-8 characters
+    utf_locale = newlocale(LC_CTYPE_MASK, "", (locale_t)0);
+    if ((locale_t)0 == utf_locale) {
+        PERROR("newlocale");
+        return 0;
+    }
+    // Capture current relevant environment variables
+    if (0 == capture_env_t(&initial_env))
+        return 0;
+    // Well done
+    return 1;
+}
+
+static void mt_safe_on_unload(void) {
+    freelocale(utf_locale);
+    release_env_t(&initial_env);
+}
+
 
 /* These are the methods we are try to hijack.
  */
 
 #ifdef HAVE_EXECVE
 int execve(const char *path, char *const argv[], char *const envp[]) {
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     return call_execve(path, argv, envp);
 }
 #endif
@@ -190,35 +228,35 @@ int execve(const char *path, char *const argv[], char *const envp[]) {
 #error can not implement execv without execve
 #endif
 int execv(const char *path, char *const argv[]) {
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     return call_execve(path, argv, environ);
 }
 #endif
 
 #ifdef HAVE_EXECVPE
 int execvpe(const char *file, char *const argv[], char *const envp[]) {
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     return call_execvpe(file, argv, envp);
 }
 #endif
 
 #ifdef HAVE_EXECVP
 int execvp(const char *file, char *const argv[]) {
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     return call_execvp(file, argv);
 }
 #endif
 
 #ifdef HAVE_EXECVP2
 int execvP(const char *file, const char *search_path, char *const argv[]) {
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     return call_execvP(file, search_path, argv);
 }
 #endif
 
 #ifdef HAVE_EXECT
 int exect(const char *path, char *const argv[], char *const envp[]) {
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     return call_exect(path, argv, envp);
 }
 #endif
@@ -230,13 +268,13 @@ int exect(const char *path, char *const argv[], char *const envp[]) {
 int execl(const char *path, const char *arg, ...) {
     va_list args;
     va_start(args, arg);
-    char const **argv = bear_strings_build(arg, &args);
+    char const **argv = string_array_from_varargs(arg, &args);
     va_end(args);
 
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     int const result = call_execve(path, (char *const *)argv, environ);
 
-    bear_strings_release(argv);
+    string_array_release(argv);
     return result;
 }
 #endif
@@ -248,13 +286,13 @@ int execl(const char *path, const char *arg, ...) {
 int execlp(const char *file, const char *arg, ...) {
     va_list args;
     va_start(args, arg);
-    char const **argv = bear_strings_build(arg, &args);
+    char const **argv = string_array_from_varargs(arg, &args);
     va_end(args);
 
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     int const result = call_execvp(file, (char *const *)argv);
 
-    bear_strings_release(argv);
+    string_array_release(argv);
     return result;
 }
 #endif
@@ -267,15 +305,15 @@ int execlp(const char *file, const char *arg, ...) {
 int execle(const char *path, const char *arg, ...) {
     va_list args;
     va_start(args, arg);
-    char const **argv = bear_strings_build(arg, &args);
+    char const **argv = string_array_from_varargs(arg, &args);
     char const **envp = va_arg(args, char const **);
     va_end(args);
 
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     int const result =
         call_execve(path, (char *const *)argv, (char *const *)envp);
 
-    bear_strings_release(argv);
+    string_array_release(argv);
     return result;
 }
 #endif
@@ -285,7 +323,7 @@ int posix_spawn(pid_t *restrict pid, const char *restrict path,
                 const posix_spawn_file_actions_t *file_actions,
                 const posix_spawnattr_t *restrict attrp,
                 char *const argv[restrict], char *const envp[restrict]) {
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     return call_posix_spawn(pid, path, file_actions, attrp, argv, envp);
 }
 #endif
@@ -295,7 +333,7 @@ int posix_spawnp(pid_t *restrict pid, const char *restrict file,
                  const posix_spawn_file_actions_t *file_actions,
                  const posix_spawnattr_t *restrict attrp,
                  char *const argv[restrict], char *const envp[restrict]) {
-    bear_report_call((char const *const *)argv);
+    report_call((char const *const *)argv);
     return call_posix_spawnp(pid, file, file_actions, attrp, argv, envp);
 }
 #endif
@@ -310,9 +348,9 @@ static int call_execve(const char *path, char *const argv[],
 
     DLSYM(func, fp, "execve");
 
-    char const **const menvp = bear_update_environment(envp, &initial_env);
+    char const **const menvp = string_array_partial_update(envp, &initial_env);
     int const result = (*fp)(path, argv, (char *const *)menvp);
-    bear_strings_release(menvp);
+    string_array_release(menvp);
     return result;
 }
 #endif
@@ -324,9 +362,9 @@ static int call_execvpe(const char *file, char *const argv[],
 
     DLSYM(func, fp, "execvpe");
 
-    char const **const menvp = bear_update_environment(envp, &initial_env);
+    char const **const menvp = string_array_partial_update(envp, &initial_env);
     int const result = (*fp)(file, argv, (char *const *)menvp);
-    bear_strings_release(menvp);
+    string_array_release(menvp);
     return result;
 }
 #endif
@@ -338,11 +376,11 @@ static int call_execvp(const char *file, char *const argv[]) {
     DLSYM(func, fp, "execvp");
 
     char **const original = environ;
-    char const **const modified = bear_update_environment(original, &initial_env);
+    char const **const modified = string_array_partial_update(original, &initial_env);
     environ = (char **)modified;
     int const result = (*fp)(file, argv);
     environ = original;
-    bear_strings_release(modified);
+    string_array_release(modified);
 
     return result;
 }
@@ -356,11 +394,11 @@ static int call_execvP(const char *file, const char *search_path,
     DLSYM(func, fp, "execvP");
 
     char **const original = environ;
-    char const **const modified = bear_update_environment(original, &initial_env);
+    char const **const modified = string_array_partial_update(original, &initial_env);
     environ = (char **)modified;
     int const result = (*fp)(file, search_path, argv);
     environ = original;
-    bear_strings_release(modified);
+    string_array_release(modified);
 
     return result;
 }
@@ -373,9 +411,9 @@ static int call_exect(const char *path, char *const argv[],
 
     DLSYM(func, fp, "exect");
 
-    char const **const menvp = bear_update_environment(envp, &initial_env);
+    char const **const menvp = string_array_partial_update(envp, &initial_env);
     int const result = (*fp)(path, argv, (char *const *)menvp);
-    bear_strings_release(menvp);
+    string_array_release(menvp);
     return result;
 }
 #endif
@@ -393,10 +431,10 @@ static int call_posix_spawn(pid_t *restrict pid, const char *restrict path,
 
     DLSYM(func, fp, "posix_spawn");
 
-    char const **const menvp = bear_update_environment(envp, &initial_env);
+    char const **const menvp = string_array_partial_update(envp, &initial_env);
     int const result =
         (*fp)(pid, path, file_actions, attrp, argv, (char *const *restrict)menvp);
-    bear_strings_release(menvp);
+    string_array_release(menvp);
     return result;
 }
 #endif
@@ -414,61 +452,54 @@ static int call_posix_spawnp(pid_t *restrict pid, const char *restrict file,
 
     DLSYM(func, fp, "posix_spawnp");
 
-    char const **const menvp = bear_update_environment(envp, &initial_env);
+    char const **const menvp = string_array_partial_update(envp, &initial_env);
     int const result =
         (*fp)(pid, file, file_actions, attrp, argv, (char *const *restrict)menvp);
-    bear_strings_release(menvp);
+    string_array_release(menvp);
     return result;
 }
 #endif
 
 /* this method is to write log about the process creation. */
 
-static void bear_report_call(char const *const argv[]) {
+static void report_call(char const *const argv[]) {
     if (!initialized)
         return;
-
-    pthread_mutex_lock(&mutex);
-    char const * const saved_locale = strdup(setlocale(LC_ALL, NULL));
-    if (0 == saved_locale) {
-        perror("bear: strdup for locale");
-        exit(EXIT_FAILURE);
-    }
-    const char *cwd = getcwd(NULL, 0);
-    if (0 == cwd) {
-        perror("bear: getcwd");
-        exit(EXIT_FAILURE);
-    }
+    // Create report file name
     char const * const out_dir = initial_env[0];
     size_t const path_max_length = strlen(out_dir) + 32;
     char filename[path_max_length];
-    if (-1 == snprintf(filename, path_max_length, "%s/execution.XXXXXX", out_dir)) {
-        perror("bear: snprintf");
-        exit(EXIT_FAILURE);
-    }
+    if (-1 == snprintf(filename, path_max_length, "%s/execution.XXXXXX", out_dir))
+        ERROR_AND_EXIT("snprintf");
+    // Create report file
     int fd = mkstemp((char *)&filename);
-    if (-1 == fd) {
-        perror("bear: mkstemp");
-        exit(EXIT_FAILURE);
-    }
-    if (0 > bear_write_json_report(fd, argv, cwd, getpid())) {
-        perror("bear: writing json problem");
-        exit(EXIT_FAILURE);
-    }
-    if (close(fd)) {
-        perror("bear: close");
-        exit(EXIT_FAILURE);
-    }
-    free((void *)cwd);
-    // Restore locale.
-    setlocale(LC_ALL, saved_locale);
-    free((void *)saved_locale);
-    pthread_mutex_unlock(&mutex);
+    if (-1 == fd)
+        ERROR_AND_EXIT("mkstemp");
+    // Write report file
+    write_report(fd, argv);
+    // Close report file
+    if (close(fd))
+        ERROR_AND_EXIT("close");
 }
 
-static int bear_write_json_report(int fd, char const *const cmd[], char const *const cwd, pid_t pid) {
-    setlocale(LC_ALL, "en_US.UTF-8");
+static void write_report(int fd, char const *const argv[]) {
+    const locale_t saved_locale = uselocale(utf_locale);
+    if ((locale_t)0 == saved_locale)
+        ERROR_AND_EXIT("uselocale");
 
+    const char *cwd = getcwd(NULL, 0);
+    if (0 == cwd)
+        ERROR_AND_EXIT("getcwd");
+    if (write_json_report(fd, argv, cwd, getpid()))
+        ERROR_AND_EXIT("writing json problem");
+    free((void *)cwd);
+
+    const locale_t restored_locale = uselocale(saved_locale);
+    if ((locale_t)0 == restored_locale)
+        ERROR_AND_EXIT("uselocale");
+}
+
+static int write_json_report(int fd, char const *const cmd[], char const *const cwd, pid_t pid) {
     if (0 > dprintf(fd, "{ \"pid\": %d, \"cmd\": [", pid))
         return -1;
 
@@ -476,14 +507,14 @@ static int bear_write_json_report(int fd, char const *const cmd[], char const *c
         char const *const sep = (it != cmd) ? "," : "";
         const size_t buffer_size = (6 * strlen(*it)) + 1;
         char buffer[buffer_size];
-        if (-1 == bear_encode_json_string(*it, buffer, buffer_size))
+        if (-1 == encode_json_string(*it, buffer, buffer_size))
             return -1;
         if (0 > dprintf(fd, "%s \"%s\"", sep, buffer))
             return -1;
     }
     const size_t buffer_size = 6 * strlen(cwd);
     char buffer[buffer_size];
-    if (-1 == bear_encode_json_string(cwd, buffer, buffer_size))
+    if (-1 == encode_json_string(cwd, buffer, buffer_size))
         return -1;
     if (0 > dprintf(fd, "], \"cwd\": \"%s\" }", buffer))
         return -1;
@@ -491,11 +522,11 @@ static int bear_write_json_report(int fd, char const *const cmd[], char const *c
     return 0;
 }
 
-static int bear_encode_json_string(char const *const src, char *const dst, size_t const dst_size) {
+static int encode_json_string(char const *const src, char *const dst, size_t const dst_size) {
     size_t const wsrc_length = mbstowcs(NULL, src, 0);
     wchar_t wsrc[wsrc_length + 1];
     if (mbstowcs((wchar_t *)&wsrc, src, wsrc_length + 1) != wsrc_length) {
-        perror("bear: mbstowcs");
+        PERROR("mbstowcs");
         return -1;
     }
     wchar_t const *wsrc_it = (wchar_t const *)&wsrc;
@@ -551,32 +582,35 @@ static int bear_encode_json_string(char const *const src, char *const dst, size_
 /* update environment assure that chilren processes will copy the desired
  * behaviour */
 
-static int bear_capture_env_t(bear_env_t *env) {
+static int capture_env_t(bear_env_t *env) {
     int status = 1;
     for (size_t it = 0; it < ENV_SIZE; ++it) {
         char const * const env_value = getenv(env_names[it]);
         char const * const env_copy = (env_value) ? strdup(env_value) : env_value;
         (*env)[it] = env_copy;
         status &= (env_copy) ? 1 : 0;
+        // Just report the problem, but don't roll back.
+        if (0 == status)
+            PERROR("strdup");
     }
     return status;
 }
 
-static void bear_release_env_t(bear_env_t *env) {
+static void release_env_t(bear_env_t *env) {
     for (size_t it = 0; it < ENV_SIZE; ++it) {
         free((void *)(*env)[it]);
         (*env)[it] = 0;
     }
 }
 
-static char const **bear_update_environment(char *const envp[], bear_env_t *env) {
-    char const **result = bear_strings_copy((char const **)envp);
+static char const **string_array_partial_update(char *const envp[], bear_env_t *env) {
+    char const **result = string_array_copy((char const **)envp);
     for (size_t it = 0; it < ENV_SIZE && (*env)[it]; ++it)
-        result = bear_update_environ(result, env_names[it], (*env)[it]);
+        result = string_array_single_update(result, env_names[it], (*env)[it]);
     return result;
 }
 
-static char const **bear_update_environ(char const *envs[], char const *key, char const * const value) {
+static char const **string_array_single_update(char const *envs[], char const *key, char const * const value) {
     // find the key if it's there
     size_t const key_length = strlen(key);
     char const **it = envs;
@@ -589,95 +623,75 @@ static char const **bear_update_environ(char const *envs[], char const *key, cha
     size_t const value_length = strlen(value);
     size_t const env_length = key_length + value_length + 2;
     char *env = malloc(env_length);
-    if (0 == env) {
-        perror("bear: malloc [in env_update]");
-        exit(EXIT_FAILURE);
-    }
-    if (-1 == snprintf(env, env_length, "%s=%s", key, value)) {
-        perror("bear: snprintf");
-        exit(EXIT_FAILURE);
-    }
+    if (0 == env)
+        ERROR_AND_EXIT("malloc");
+    if (-1 == snprintf(env, env_length, "%s=%s", key, value))
+        ERROR_AND_EXIT("snprintf");
     // replace or append the environment entry
     if (it && *it) {
         free((void *)*it);
         *it = env;
-	return envs;
+	    return envs;
+    } else {
+        size_t const size = string_array_length(envs);
+        char const **result = realloc(envs, (size + 2) * sizeof(char const *));
+        if (0 == result)
+            ERROR_AND_EXIT("realloc");
+        result[size] = env;
+        result[size + 1] = 0;
+        return result;
     }
-    return bear_strings_append(envs, env);
 }
 
 /* util methods to deal with string arrays. environment and process arguments
  * are both represented as string arrays. */
 
-static char const **bear_strings_build(char const *const arg, va_list *args) {
+static char const **string_array_from_varargs(char const *const arg, va_list *args) {
     char const **result = 0;
     size_t size = 0;
     for (char const *it = arg; it; it = va_arg(*args, char const *)) {
         result = realloc(result, (size + 1) * sizeof(char const *));
-        if (0 == result) {
-            perror("bear: realloc");
-            exit(EXIT_FAILURE);
-        }
+        if (0 == result)
+            ERROR_AND_EXIT("realloc");
         char const *copy = strdup(it);
-        if (0 == copy) {
-            perror("bear: strdup");
-            exit(EXIT_FAILURE);
-        }
+        if (0 == copy)
+            ERROR_AND_EXIT("strdup");
         result[size++] = copy;
     }
     result = realloc(result, (size + 1) * sizeof(char const *));
-    if (0 == result) {
-        perror("bear: realloc");
-        exit(EXIT_FAILURE);
-    }
+    if (0 == result)
+        ERROR_AND_EXIT("realloc");
     result[size++] = 0;
 
     return result;
 }
 
-static char const **bear_strings_copy(char const **const in) {
-    size_t const size = bear_strings_length(in);
+static char const **string_array_copy(char const **const in) {
+    size_t const size = string_array_length(in);
 
     char const **const result = malloc((size + 1) * sizeof(char const *));
-    if (0 == result) {
-        perror("bear: malloc");
-        exit(EXIT_FAILURE);
-    }
+    if (0 == result)
+        ERROR_AND_EXIT("malloc");
 
     char const **out_it = result;
     for (char const *const *in_it = in; (in_it) && (*in_it);
          ++in_it, ++out_it) {
         *out_it = strdup(*in_it);
-        if (0 == *out_it) {
-            perror("bear: strdup");
-            exit(EXIT_FAILURE);
-        }
+        if (0 == *out_it)
+            ERROR_AND_EXIT("strdup");
     }
     *out_it = 0;
     return result;
 }
 
-static char const **bear_strings_append(char const **const in,
-                                        char const *const e) {
-    size_t size = bear_strings_length(in);
-    char const **result = realloc(in, (size + 2) * sizeof(char const *));
-    if (0 == result) {
-        perror("bear: realloc");
-        exit(EXIT_FAILURE);
-    }
-    result[size++] = e;
-    result[size++] = 0;
-    return result;
-}
-
-static size_t bear_strings_length(char const *const *const in) {
+static size_t string_array_length(char const *const *const in) {
     size_t result = 0;
     for (char const *const *it = in; (it) && (*it); ++it)
         ++result;
     return result;
 }
 
-static void bear_strings_release(char const **in) {
+static void string_array_release(char const **in) {
     for (char const *const *it = in; (it) && (*it); ++it) {
         free((void *)*it);
     }
